import unittest
from opentc.server import Server
from opentc.client import Client
import json
import tempfile
import os
import time
import logging
from opentc import setup_logging
from opentc import __version__


class TestOpentc(unittest.TestCase):
    config_directories = [os.curdir, os.path.expanduser("~/.opentc"), "/etc/opentc", os.environ.get("OPENTC_CONF_DIR")]
    setup_logging(config_directories=config_directories)
    logger = logging.getLogger(__name__)
    run_server = False
    address, port = "localhost", 3333
    server = None
    tcc = None
    data = b"Text Classification"*1000
    # md5sum = md5(data).hexdigest()
    md5sum = "4b1c78bb298ef3d3d3ee9a244cb5e0c6"
    x_raw = ["We would like to convert a group of these outlines into a 3D image. Someone mentioned that if we could \
                convert the TIFF into a vector format then we could view them in Autocad",
             "We do baptize converts, but no one who has been deceived into hearing the word is likely to be \
                a convert. If in fact the grace of God might work in such a situation, there is no harm done in \
                waiting a day or two",
             "Chronic persistent hepatitis is usually diagnosed when someone does a liver biopsy on a patient \
                that has persistently elevated serum transaminases months after a bout of acute viral hepatitis"]

    @classmethod
    def setUpClass(cls):
        TestOpentc.logger.debug("setUpClass")
        cls.tcc = Client(address=cls.address, port=cls.port)
        if cls.run_server:
            cls.server = Server()
            cls.server.start(run_forever=False, address=cls.address, port=cls.port)

    @classmethod
    def tearDownClass(cls):
        TestOpentc.logger.debug("tearDownClass")
        if cls.run_server:
            cls.server.shutdown()

    def setUp(self):
        self.logger.debug("setUp")

    def tearDown(self):
        self.logger.debug("tearDown")

    def test_ping(self):
        response = self.tcc.command("PING\n")
        response = json.loads(response.decode('utf-8'))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual('PONG', response['result'])

    def test_version(self):
        response = self.tcc.command("VERSION\n")
        response = json.loads(response.decode('utf-8'))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual(__version__, response['result'])

    def test_list_classifier(self):
        response = self.tcc.command("LIST_CLASSIFIER\n")
        response = json.loads(response.decode('utf-8'))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual({'cnn': True, 'bayes': True, 'svm': True}, response['result'])

    def test_set_classifier(self):
        response = self.tcc.command("SET_CLASSIFIER:bayes:False\n")
        response = json.loads(response.decode('utf-8'))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual([{'bayes': False}], response['result'])
        response = self.tcc.command("SET_CLASSIFIER:bayes:True\n")
        response = json.loads(response.decode('utf-8'))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual([{'bayes': True}], response['result'])

    def test_md5_file(self):
        fd, temp_path = tempfile.mkstemp()
        file = os.fdopen(fd, "wb")
        file.write(self.data)
        file.close()
        response = self.tcc.md5_file(temp_path)
        response = json.loads(response.decode('utf-8'))
        os.remove(temp_path)
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual(self.md5sum, response['result'])

    def test_md5_stream(self):
        response = self.tcc.md5_stream(self.data)
        response = json.loads(response.decode('utf-8'))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual(self.md5sum, response['result'])

    def test_predict_stream_0(self):
        start = time.time()
        response = self.tcc.predict_stream(self.x_raw[0].encode('utf-8'))
        response = json.loads(response.decode('utf-8'))
        end = time.time()
        self.logger.debug("Time elapsed: {}".format(end - start))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual({'bayes': ['comp.graphics'],
                          'cnn': ['comp.graphics'],
                          'svm': ['comp.graphics']}, response['result'])

    def test_predict_stream_1(self):
        start = time.time()
        response = self.tcc.predict_stream(self.x_raw[1].encode('utf-8'))
        response = json.loads(response.decode('utf-8'))
        end = time.time()
        self.logger.debug("Time elapsed: {}".format(end - start))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual({'bayes': ['soc.religion.christian'],
                          'cnn': ['soc.religion.christian'],
                          'svm': ['soc.religion.christian']}, response['result'])

    def test_predict_stream_2(self):
        start = time.time()
        response = self.tcc.predict_stream('\n'.join(self.x_raw).encode('utf-8'))
        response = json.loads(response.decode('utf-8'))
        end = time.time()
        self.logger.debug("Time elapsed: {}".format(end - start))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual({'bayes': ['comp.graphics', 'soc.religion.christian', 'sci.med'],
                          'cnn': ['comp.graphics', 'soc.religion.christian', 'sci.med'],
                          'svm': ['comp.graphics', 'soc.religion.christian', 'sci.med']},
                         response['result'])

    def test_predict_file(self):
        fd, temp_path = tempfile.mkstemp()
        file = os.fdopen(fd, "wb")
        file.write(self.x_raw[1].encode('utf-8'))
        file.close()
        response = self.tcc.predict_file(temp_path)
        response = json.loads(response.decode('utf-8'))
        os.remove(temp_path)
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual({'bayes': ['soc.religion.christian'],
                          'cnn': ['soc.religion.christian'],
                          'svm': ['soc.religion.christian']}, response['result'])

    def test_predict_file_multilines(self):
        fd, temp_path = tempfile.mkstemp()
        file = os.fdopen(fd, "wb")
        file.write('\n'.join(self.x_raw).encode('utf-8'))
        file.close()
        response = self.tcc.predict_file(temp_path)
        response = json.loads(response.decode('utf-8'))
        os.remove(temp_path)
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual({'bayes': ['comp.graphics', 'soc.religion.christian', 'sci.med'],
                          'cnn': ['comp.graphics', 'soc.religion.christian', 'sci.med'],
                          'svm': ['comp.graphics', 'soc.religion.christian', 'sci.med']},
                         response['result'])

    def test_unknown_command(self):
        response = self.tcc.command("Unknown command\n")
        response = json.loads(response.decode('utf-8'))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual('Unknown Command', response['result'])

    def test_close(self):
        response = self.tcc.command("CLOSE\n")
        response = json.loads(response.decode('utf-8'))
        self.logger.debug("{}: {}".format(self._testMethodName, response))
        self.assertEqual('Bye', response['result'])
        TestOpentc.tcc = Client(address=self.address, port=self.port)

if __name__ == '__main__':
    unittest.main(verbosity=2)



from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.pipeline import Pipeline
from sklearn.linear_model import SGDClassifier
from sklearn.externals import joblib
from . import Classifier


class Svm(Classifier):

    def __init__(self, cfg=None, categories=None, current_category=None, load=True):
        super().__init__()
        self.categories = categories
        self.current_category = current_category
        if load:
            self.clf = joblib.load(cfg['pre_trained_file'][self.current_category])

    def fit(self, dataset, filename):
        self.logger.debug("fit")
        self.clf = Pipeline([('vect', CountVectorizer()),
                             ('tfidf', TfidfTransformer()),
                             ('clf', SGDClassifier(loss='hinge', penalty='l2', alpha=1e-3, n_iter=5, random_state=42)),
                             ])
        self.clf.fit(dataset.get_dataset()['data'], dataset.get_dataset()['target'])
        joblib.dump(self.clf, filename + ".pkl", compress=9)

    def reload(self, filename):
        self.logger.info("reload")
        self.clf = joblib.load(filename)

    def predict(self, data):
        self.logger.debug("predict")
        predicted = self.clf.predict(data)
        predicted = [self.categories[i] for i in predicted]
        return predicted



import logging
from abc import ABC, abstractmethod


class Classifier(ABC):
    __instances__ = dict()

    def __init__(self):
        self.logger = logging.getLogger(self.__module__)
        Classifier.__instances__[self.__class__.__name__] = self

    @abstractmethod
    def fit(self, dataset, filename):
        pass

    @abstractmethod
    def reload(self, filename):
        pass

    @abstractmethod
    def predict(self, data):
        pass


import os
import logging
from unittest import TestCase
from opentc import setup_logging


class TestSetupLogging(TestCase):
    def test_logging_print(self):
        config_directories = [os.curdir, os.path.expanduser("~/.opentc"), "/etc/opentc", os.environ.get("OPENTC_CONF_DIR")]
        setup_logging(config_directories=config_directories)
        logger = logging.getLogger(__name__)
        logger.info("logging works")
        self.assertTrue(logging is not None)


from . cnn_text import CnnTextTraining, CnnTextEvaluator
from . import Classifier


class Cnn(Classifier):

    def __init__(self, cfg=None, categories=None, current_category=None, load=True):
        super().__init__()
        self.cfg = cfg
        self.categories = categories
        self.current_category = current_category
        if load:
            self.evaluator = CnnTextEvaluator(self.cfg, self.current_category)
        self.clf = None

    def fit(self, dataset, filename):
        self.logger.debug("train")
        self.clf = CnnTextTraining(self.cfg)
        self.clf.fit(dataset, filename)

    def reload(self, filename):
        self.logger.debug("reload")

    def predict(self, data):
        self.logger.debug("predict")
        predicted = self.evaluator.predict(data)
        predicted = [self.categories[i] for i in predicted]
        return predicted



import os
import sys
import logging
import socket
import struct


class SimpleSocket(object):
    max_data_size = 4096

    def __init__(self, address=None, port=None):
        self.logger = logging.getLogger(__name__)
        self._timeout = None
        self._address = address
        self._port = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.socket.connect((self._address, self._port))
        except ConnectionError as err:
            self.logger.error("{} to {}:{}".format(err, self._address, self._port, err))
            sys.exit()

    def close(self):
        self.logger.debug("Closing main socket")
        self._close_connection()

    def _close_connection(self):
        self.socket.close()

    def send(self, data):
        size = len(data)
        packed_header = struct.pack('=I', size)
        self.socket.sendall(packed_header + data)

    def receive(self):
        packed_header = self.socket.recv(4)
        (size, ) = struct.unpack('=I', packed_header)
        if size == 0 or size > self.max_data_size:
            return None
        data = self.socket.recv(size)
        return data


class Client(object):
    max_chunk_size = 1024

    def __init__(self, address='localhost', port=3333):
        self.address = address
        self.port = port
        self.simple_socket = SimpleSocket(address=self.address, port=self.port)

    def command(self, message=None):
        logger = logging.getLogger(__name__)
        try:
            self.simple_socket.send(message.encode('utf-8'))
            response = self.simple_socket.receive()
            logger.debug("command Received: {}".format(response))
            return response
        except ConnectionError as err:
            logger.error("OS error: {0}".format(err))

    def md5_file(self, file_name=None):
        # This function is just for testing purpose
        logger = logging.getLogger(__name__)
        try:
            statinfo = os.stat(file_name)
            if statinfo is not None:
                command = "MD5_FILE:{}\n".format(file_name)
                self.simple_socket.send(command.encode('utf-8'))
                response = self.simple_socket.receive()
                logger.debug("Client Received: {}".format(response))
                return response
            else:
                return None
        except OSError as err:
            logger.error("OS error: {0}".format(err))

    def md5_stream(self, data=None):
        # This function is just for testing purpose
        logger = logging.getLogger(__name__)
        command = "MD5_STREAM\n"
        self.simple_socket.send(command.encode('utf-8'))
        try:
            data_len = len(data)
            start_pos = 0
            end_pos = self.max_chunk_size
            while start_pos < data_len:
                end_pos = min(end_pos, data_len)
                self.simple_socket.send(data[start_pos:end_pos])
                start_pos += self.max_chunk_size
                end_pos += self.max_chunk_size
            self.simple_socket.send(b'')
            response = self.simple_socket.receive()
            logger.debug("Client Received: {}".format(response))
            return response
        except OSError as err:
            logger.error("OS error: {0}".format(err))

    def predict_stream(self, data=None):
        logger = logging.getLogger(__name__)
        command = "PREDICT_STREAM\n"
        self.simple_socket.send(command.encode('utf-8'))
        try:
            data_len = len(data)
            start_pos = 0
            end_pos = self.max_chunk_size
            while start_pos < data_len:
                end_pos = min(end_pos, data_len)
                self.simple_socket.send(data[start_pos:end_pos])
                start_pos += self.max_chunk_size
                end_pos += self.max_chunk_size
            self.simple_socket.send(b'')
            response = self.simple_socket.receive()
            logger.debug("Client Received: {}".format(response))
            return response
        except OSError as err:
            logger.error("OS error: {0}".format(err))

    def predict_file(self, file_name=None):
        logger = logging.getLogger(__name__)
        try:
            statinfo = os.stat(file_name)
            if statinfo is not None:
                command = "PREDICT_FILE:{}\n".format(file_name)
                self.simple_socket.send(command.encode('utf-8'))
                response = self.simple_socket.receive()
                logger.debug("Client Received: {}".format(response))
                return response
            else:
                return None
        except OSError as err:
            logger.error("OS error: {0}".format(err))

    def set_classifier(self, classifier=None, value=None):
        logger = logging.getLogger(__name__)
        command = "SET_CLASSIFIER:{}:{}\n".format(classifier, value)
        self.simple_socket.send(command.encode('utf-8'))
        response = self.simple_socket.receive()
        logger.debug("Client Received: {}".format(response))
        return response




from sklearn.datasets import fetch_20newsgroups
from . import Dataset


class TwentyNewsgroup(Dataset):

    def __init__(self, cfg=None):
        super().__init__()
        self.__dataset__ = fetch_20newsgroups(subset=cfg['subset'], categories=cfg['categories'],
                                              shuffle=cfg['shuffle'], random_state=cfg['random_state'])




from setuptools import setup, find_packages
import opentc

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except ImportError:
    long_description = open('README.md').read()


def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name='opentc',
    version=opentc.__version__,
    description='A text classification engine using machine learning and designed as client-server architecture',
    long_description=long_description,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Text Processing :: Filters'
    ],
    keywords='machine learning cnn svm bayesian',
    url='https://github.com/cahya-wirawan/opentc',
    author='Cahya Wirawan',
    author_email='Cahya.Wirawan@gmail.com',
    license='MIT',
    packages=find_packages('.'),
    package_dir={'': '.'},
    install_requires=[
        'numpy',
        'pyparsing',
        'PyYAML',
        'scikit-learn',
        'scipy',
        'tensorflow'
    ],
    scripts=['bin/opentc', 'bin/opentcd'],
    data_files=[('/etc/opentc', ['config/opentc.yml', 'config/logging.yml'])],
    include_package_data=True,
    zip_safe=False)


"""
OpenTC is text classification engine running as client server architecture. It listen on port 3333

"""
import os
import logging.config
import yaml
import logging


__author__ = "Cahya Wirawan <Cahya.Wirawan@gmail.com>"
__version__ = '0.3.11'


def setup_logging(
        config_directories=None,
        config_file=None,
        default_level=logging.INFO
):
    """Setup logging configuration

    """
    config_found = False
    config_file_path = None
    if config_file:
        config_file_path = config_file
        if os.path.isfile(config_file_path) and os.access(config_file_path, os.R_OK):
            config_found = True
    else:
        for directory in config_directories:
            if directory is None:
                continue
            config_file_path = os.path.join(directory, "logging.yml")
            if os.path.isfile(config_file_path) and os.access(config_file_path, os.R_OK):
                config_found = True
                break
    if config_found:
        with open(config_file_path, 'rt') as ymlfile:
            config = yaml.safe_load(ymlfile.read())
        logging.config.dictConfig(config)
    else:
        logging.basicConfig(level=default_level)


def setup_config(
        config_directories=None,
        config_file=None,
        default_level=logging.INFO
):
    """Setup logging configuration

    """
    config_found = False
    config_file_path = None
    if config_file:
        config_file_path = config_file
        if os.path.isfile(config_file_path) and os.access(config_file_path, os.R_OK):
            config_found = True
    else:
        for directory in config_directories:
            if directory is None:
                continue
            config_file_path = os.path.join(directory, "opentc.yml")
            if os.path.isfile(config_file_path) and os.access(config_file_path, os.R_OK):
                config_found = True
                break
    if config_found:
        with open(config_file_path, 'rt') as ymlfile:
            config = yaml.safe_load(ymlfile.read())
        return config
    else:
        print("The configuration file is not found.")
        exit(1)



import numpy as np
import re
import logging
from abc import ABC, abstractmethod


class Dataset(ABC):
    __dataset__ = None
    __instances__ = dict()

    def __init__(self):
        self.logger = logging.getLogger(self.__module__)
        Dataset.__instances__[self.__class__.__name__] = self

    @staticmethod
    def clean_string(string):
        """
        Tokenization/string cleaning for all datasets except for SST.
        Original taken from https://github.com/yoonkim/CNN_sentence/blob/master/process_data.py
        """
        string = re.sub(r"[^A-Za-z0-9(),!?\'\`]", " ", string)
        string = re.sub(r"\'s", " \'s", string)
        string = re.sub(r"\'ve", " \'ve", string)
        string = re.sub(r"n\'t", " n\'t", string)
        string = re.sub(r"\'re", " \'re", string)
        string = re.sub(r"\'d", " \'d", string)
        string = re.sub(r"\'ll", " \'ll", string)
        string = re.sub(r",", " , ", string)
        string = re.sub(r"!", " ! ", string)
        string = re.sub(r"\(", " \( ", string)
        string = re.sub(r"\)", " \) ", string)
        string = re.sub(r"\?", " \? ", string)
        string = re.sub(r"\s{2,}", " ", string)
        return string.strip().lower()

    def load_data_labels(self):
        """
        Load data and labels
        :return: data and labels
        """
        self.logger.debug("load_data_labels")
        # Split by words
        x_text = self.__dataset__['data']
        x_text = [Dataset.clean_string(sent) for sent in x_text]
        # Generate labels
        labels = []
        for i in range(len(x_text)):
            label = [0 for j in self.__dataset__['target_names']]
            label[self.__dataset__['target'][i]] = 1
            labels.append(label)
        y = np.array(labels)
        return [x_text, y]

    def get_dataset(self):
        return self.__dataset__

    @staticmethod
    def create_dataset(dataset):
        module = __import__("opentc.dataset." + dataset['name'])
        class_ = getattr(getattr(getattr(module, "dataset"), dataset['name']), dataset['name'].title().replace("_",""))
        instance = class_(dataset)
        return instance




import os
import logging.config
import logging
from datetime import datetime
from opentc.dataset import Dataset


class Training(object):
    """
    Class for using OpenTCServer with a network socket
    """
    classifiers = dict()

    def __init__(self, cfg):
        """
        class initialisation
        """
        self.logger = logging.getLogger(__name__)
        self.cfg = cfg

        try:
            os.makedirs(self.cfg["result_dir"], exist_ok=True)
        except OSError as err:
            self.logger.error("OS error: {0}".format(err))

        for classifier_name in self.cfg["classifiers"]:
            if classifier_name == "default":
                continue
            module = __import__("opentc.classifier." + classifier_name)
            class_ = getattr(getattr(getattr(module, "classifier"), classifier_name), classifier_name.title())
            if class_:
                classifier = dict()
                classifier["enabled"] = self.cfg["classifiers"][classifier_name]["enabled"]
                default_dataset = self.cfg["datasets"]["default"]
                classifier["class"] = class_(self.cfg["classifiers"][classifier_name],
                                             self.cfg["datasets"][default_dataset]["categories"],
                                             default_dataset, False)
                Training.classifiers[classifier_name] = classifier

    def start(self, cn=None, dn=None):
        self.logger.info("Training starts")
        if cn:
            classifier_name = cn
        else:
            classifier_name = self.cfg["classifiers"]["default"]
        if classifier_name not in Training.classifiers.keys() and classifier_name != "all":
            print("The classifier {} doesn't exist".format(classifier_name))
            return 1
        if dn:
            dataset_name = dn
        else:
            dataset_name = self.cfg["datasets"]["default"]
        if dataset_name not in self.cfg["datasets"]:
            print("The dataset {} doesn't exist".format(dataset_name))
            return 1
        dataset = Dataset.create_dataset(self.cfg["datasets"][dataset_name])
        now = datetime.now().strftime('%Y%m%d_%H%M%S')
        training = Training
        if classifier_name == "all":
            for classifier_name in training.classifiers:
                result_name = "{}/{}_{}_{}".format(self.cfg["result_dir"], classifier_name, dataset_name, now)
                training.classifiers[classifier_name]["class"].fit(dataset, result_name)
                print("The training of {} classifier for the dataset {} is done.".format(classifier_name, dataset_name))
                print("The result is saved in: {}(.pkl)".format(result_name))
        else:
            result_name = "{}/{}_{}_{}".format(self.cfg["result_dir"], classifier_name, dataset_name, now)
            training.classifiers[classifier_name]["class"].fit(dataset, result_name)
            print("The training of {} classifier for the dataset {} is done.".format(classifier_name, dataset_name))
            print("The result is saved in: {}(.pkl)".format(result_name))

        self.logger.info("Training end")
